/**************************************************************************

 Written by Imran S. Haque (ihaque@cs.stanford.edu)

 Copyright (c) 2011 Stanford University.
 All rights reserved.

 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.
     * Neither the name of Stanford University nor the names of its contributors
       may be used to endorse or promote products derived from this software without
       specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

**************************************************************************/

#ifdef _DEBUG
#include <assert.h>
#else
#define assert(x)
#endif

#include <emmintrin.h>
#include <stdint.h>

inline __m128i reduce_16_to_32(__m128i v) {
    // Reduce 8*16->4*32
    const __m128i ZERO = _mm_setzero_si128();
    __m128i temp1 = _mm_unpacklo_epi16(v,ZERO);
    __m128i temp2 = _mm_unpackhi_epi16(v,ZERO);
    return _mm_add_epi32(temp1,temp2);
    //const __m128i ONE = _mm_set1_epi16(0x0001);
    //return _mm_madd_epi16(v,ONE);

}
inline __m128i popcount_sse2_16_helper_1(unsigned* buf, int N) {
        __m128i* vbuf = (__m128i*)buf;
        __m128i total,count0,count1,count2,count3;
        int i;
        __m128i B0 = _mm_set1_epi32(0x55555555);
        __m128i B1 = _mm_set1_epi32(0x33333333);
        __m128i B2 = _mm_set1_epi32(0x0F0F0F0F);
        __m128i B3 = _mm_set1_epi32(0x00FF00FF);
        total = _mm_setzero_si128();
        for (i = 0; i < N; i+=4) {
            __m128i v0 = _mm_load_si128(vbuf+i+0);
            __m128i v1 = _mm_load_si128(vbuf+i+1);
            __m128i v2 = _mm_load_si128(vbuf+i+2);
            __m128i v3 = _mm_load_si128(vbuf+i+3);
            
            count0 = _mm_sub_epi16(v0, _mm_and_si128(B0, _mm_srli_epi16(v0,1)));
            count1 = _mm_sub_epi16(v1, _mm_and_si128(B0, _mm_srli_epi16(v1,1)));
            count2 = _mm_sub_epi16(v2, _mm_and_si128(B0, _mm_srli_epi16(v2,1)));
            count3 = _mm_sub_epi16(v3, _mm_and_si128(B0, _mm_srli_epi16(v3,1)));

            count0 = _mm_add_epi16(_mm_and_si128(B1,count0),
                                   _mm_and_si128(B1,_mm_srli_epi16(count0,2)));
            count1 = _mm_add_epi16(_mm_and_si128(B1,count1),
                                   _mm_and_si128(B1,_mm_srli_epi16(count1,2)));
            count2 = _mm_add_epi16(_mm_and_si128(B1,count2),
                                   _mm_and_si128(B1,_mm_srli_epi16(count2,2)));
            count3 = _mm_add_epi16(_mm_and_si128(B1,count3),
                                   _mm_and_si128(B1,_mm_srli_epi16(count3,2)));

            count0 = _mm_and_si128(B2,_mm_add_epi16(count0,
                                                    _mm_srli_epi16(count0,4)));
            count1 = _mm_and_si128(B2,_mm_add_epi16(count1,
                                                    _mm_srli_epi16(count1,4)));
            count2 = _mm_and_si128(B2,_mm_add_epi16(count2,
                                                    _mm_srli_epi16(count2,4)));
            count3 = _mm_and_si128(B2,_mm_add_epi16(count3,
                                                    _mm_srli_epi16(count3,4)));

            count0 = _mm_and_si128(B3,_mm_add_epi16(count0,
                                                   _mm_srli_epi16(count0,8)));
            count1 = _mm_and_si128(B3,_mm_add_epi16(count1,
                                                   _mm_srli_epi16(count1,8)));
            count2 = _mm_and_si128(B3,_mm_add_epi16(count2,
                                                   _mm_srli_epi16(count2,8)));
            count3 = _mm_and_si128(B3,_mm_add_epi16(count3,
                                                   _mm_srli_epi16(count3,8)));

            total = _mm_add_epi16(total,_mm_add_epi16(_mm_add_epi16(count0,count1),
                                                      _mm_add_epi16(count2,count3)));

        }

        return reduce_16_to_32(total);
}
inline int popcount_sse2_16_1(unsigned* buf,int n) {
    int N = n/4, base = 0;
    uint32_t tmp[4];
    __m128i count32 = _mm_setzero_si128();
    // 2^12 loop iters might overflow 16-bit counter, so
    // cap it at 2^11 iters per chunk
    const int maxiters = 2048;
    while (N > maxiters) {
        count32 = _mm_add_epi32(count32,popcount_sse2_16_helper_1(buf+base,maxiters));
        base += maxiters*4;
        N -= maxiters;
    }
    if (N > 0) count32 = _mm_add_epi32(count32,popcount_sse2_16_helper_1(buf+base, N));
    _mm_store_si128((__m128i*)tmp,count32);
    return tmp[0]+tmp[1]+tmp[2]+tmp[3];
}
inline __m128i popcount_sse2_16_helper_2(unsigned* bufA,unsigned* bufB, int N) {
        __m128i* vbufA = (__m128i*)bufA;
        __m128i* vbufB = (__m128i*)bufB;
        __m128i total,count0,count1,count2,count3;
        int i;
        __m128i B0 = _mm_set1_epi32(0x55555555);
        __m128i B1 = _mm_set1_epi32(0x33333333);
        __m128i B2 = _mm_set1_epi32(0x0F0F0F0F);
        __m128i B3 = _mm_set1_epi32(0x00FF00FF);
        total = _mm_setzero_si128();
        for (i = 0; i < N; i+=4) {
            __m128i v0 = _mm_and_si128(_mm_load_si128(vbufA+i+0),_mm_load_si128(vbufB+i+0));
            __m128i v1 = _mm_and_si128(_mm_load_si128(vbufA+i+1),_mm_load_si128(vbufB+i+1));
            __m128i v2 = _mm_and_si128(_mm_load_si128(vbufA+i+2),_mm_load_si128(vbufB+i+2));
            __m128i v3 = _mm_and_si128(_mm_load_si128(vbufA+i+3),_mm_load_si128(vbufB+i+3));
            
            count0 = _mm_sub_epi16(v0, _mm_and_si128(B0, _mm_srli_epi16(v0,1)));
            count1 = _mm_sub_epi16(v1, _mm_and_si128(B0, _mm_srli_epi16(v1,1)));
            count2 = _mm_sub_epi16(v2, _mm_and_si128(B0, _mm_srli_epi16(v2,1)));
            count3 = _mm_sub_epi16(v3, _mm_and_si128(B0, _mm_srli_epi16(v3,1)));

            count0 = _mm_add_epi16(_mm_and_si128(B1,count0),
                                   _mm_and_si128(B1,_mm_srli_epi16(count0,2)));
            count1 = _mm_add_epi16(_mm_and_si128(B1,count1),
                                   _mm_and_si128(B1,_mm_srli_epi16(count1,2)));
            count2 = _mm_add_epi16(_mm_and_si128(B1,count2),
                                   _mm_and_si128(B1,_mm_srli_epi16(count2,2)));
            count3 = _mm_add_epi16(_mm_and_si128(B1,count3),
                                   _mm_and_si128(B1,_mm_srli_epi16(count3,2)));

            count0 = _mm_and_si128(B2,_mm_add_epi16(count0,
                                                    _mm_srli_epi16(count0,4)));
            count1 = _mm_and_si128(B2,_mm_add_epi16(count1,
                                                    _mm_srli_epi16(count1,4)));
            count2 = _mm_and_si128(B2,_mm_add_epi16(count2,
                                                    _mm_srli_epi16(count2,4)));
            count3 = _mm_and_si128(B2,_mm_add_epi16(count3,
                                                    _mm_srli_epi16(count3,4)));

            count0 = _mm_and_si128(B3,_mm_add_epi16(count0,
                                                   _mm_srli_epi16(count0,8)));
            count1 = _mm_and_si128(B3,_mm_add_epi16(count1,
                                                   _mm_srli_epi16(count1,8)));
            count2 = _mm_and_si128(B3,_mm_add_epi16(count2,
                                                   _mm_srli_epi16(count2,8)));
            count3 = _mm_and_si128(B3,_mm_add_epi16(count3,
                                                   _mm_srli_epi16(count3,8)));

            total = _mm_add_epi16(total,_mm_add_epi16(_mm_add_epi16(count0,count1),
                                                      _mm_add_epi16(count2,count3)));

        }

        return reduce_16_to_32(total);
}
inline int popcount_sse2_16_2(unsigned* bufA,unsigned* bufB,int n) {
    int N = n/4, base = 0;
    uint32_t tmp[4];
    __m128i count32 = _mm_setzero_si128();
    // 2^12 loop iters might overflow 16-bit counter, so
    // cap it at 2^11 iters per chunk
    const int maxiters = 2048;
    while (N > maxiters) {
        count32 = _mm_add_epi32(count32,popcount_sse2_16_helper_2(bufA+base,bufB+base,maxiters));
        base += maxiters*4;
        N -= maxiters;
    }
    if (N > 0) count32 = _mm_add_epi32(count32,popcount_sse2_16_helper_2(bufA+base, bufB+base, N));
    _mm_store_si128((__m128i*)tmp,count32);
    return tmp[0]+tmp[1]+tmp[2]+tmp[3];
}
inline void popcount_sse2_16_helper_2x1(const unsigned* bufR0, const unsigned* bufR1, const unsigned* bufB, const int N,__m128i& countR0,__m128i& countR1) {
        __m128i* vbufR0 = (__m128i*)bufR0;
        __m128i* vbufR1 = (__m128i*)bufR1;
        __m128i* vbufB = (__m128i*)bufB;
        __m128i totalR0,totalR1,countR0_0,countR0_1,countR1_0,countR1_1;
        const __m128i M0 = _mm_set1_epi32(0x55555555);
        const __m128i M1 = _mm_set1_epi32(0x33333333);
        const __m128i M2 = _mm_set1_epi32(0x0F0F0F0F);
        const __m128i M3 = _mm_set1_epi32(0x00FF00FF);
        __m128i vB0,vB1;
        totalR0 = _mm_setzero_si128();
        totalR1 = _mm_setzero_si128();
        for (int i = 0; i < N; i+=2) {
                vB0 = _mm_load_si128(vbufB+i);
                vB1 = _mm_load_si128(vbufB+i+1);
                __m128i vR0_0 = _mm_and_si128(_mm_load_si128(vbufR0+i+0),vB0);
                __m128i vR0_1 = _mm_and_si128(_mm_load_si128(vbufR0+i+1),vB1);
                __m128i vR1_0 = _mm_and_si128(_mm_load_si128(vbufR1+i+0),vB0);
                __m128i vR1_1 = _mm_and_si128(_mm_load_si128(vbufR1+i+1),vB1);
                
                countR0_0 = _mm_sub_epi16(vR0_0, _mm_and_si128(M0, _mm_srli_epi16(vR0_0,1)));
                countR0_1 = _mm_sub_epi16(vR0_1, _mm_and_si128(M0, _mm_srli_epi16(vR0_1,1)));
                countR1_0 = _mm_sub_epi16(vR1_0, _mm_and_si128(M0, _mm_srli_epi16(vR1_0,1)));
                countR1_1 = _mm_sub_epi16(vR1_1, _mm_and_si128(M0, _mm_srli_epi16(vR1_1,1)));

                countR0_0 = _mm_add_epi16(_mm_and_si128(M1,countR0_0),
                                          _mm_and_si128(M1,_mm_srli_epi16(countR0_0,2)));
                countR0_1 = _mm_add_epi16(_mm_and_si128(M1,countR0_1),
                                          _mm_and_si128(M1,_mm_srli_epi16(countR0_1,2)));
                countR1_0 = _mm_add_epi16(_mm_and_si128(M1,countR1_0),
                                          _mm_and_si128(M1,_mm_srli_epi16(countR1_0,2)));
                countR1_1 = _mm_add_epi16(_mm_and_si128(M1,countR1_1),
                                          _mm_and_si128(M1,_mm_srli_epi16(countR1_1,2)));

                countR0_0 = _mm_and_si128(M2,_mm_add_epi16(countR0_0,
                                                           _mm_srli_epi16(countR0_0,4)));
                countR0_1 = _mm_and_si128(M2,_mm_add_epi16(countR0_1,
                                                           _mm_srli_epi16(countR0_1,4)));
                countR1_0 = _mm_and_si128(M2,_mm_add_epi16(countR1_0,
                                                           _mm_srli_epi16(countR1_0,4)));
                countR1_1 = _mm_and_si128(M2,_mm_add_epi16(countR1_1,
                                                           _mm_srli_epi16(countR1_1,4)));

                countR0_0 = _mm_and_si128(M3,_mm_add_epi16(countR0_0,
                                                       _mm_srli_epi16(countR0_0,8)));
                countR0_1 = _mm_and_si128(M3,_mm_add_epi16(countR0_1,
                                                       _mm_srli_epi16(countR0_1,8)));
                countR1_0 = _mm_and_si128(M3,_mm_add_epi16(countR1_0,
                                                       _mm_srli_epi16(countR1_0,8)));
                countR1_1 = _mm_and_si128(M3,_mm_add_epi16(countR1_1,
                                                       _mm_srli_epi16(countR1_1,8)));

                totalR0 = _mm_add_epi16(totalR0,_mm_add_epi16(countR0_0,countR0_1));
                totalR1 = _mm_add_epi16(totalR1,_mm_add_epi16(countR1_0,countR1_1));
        }
        // Partial horizontal add-reduction
        countR0 = _mm_add_epi32(countR0,reduce_16_to_32(totalR0));
        countR1 = _mm_add_epi32(countR1,reduce_16_to_32(totalR1));
        return;
}
inline void popcount_sse2_16_2x1(const unsigned* bufR0,const unsigned* bufR1,const unsigned* bufB,int n,int* counts) {
    int N = n/4, base = 0;
    uint32_t tmp[4];
    __m128i count32_R0 = _mm_setzero_si128();
    __m128i count32_R1 = _mm_setzero_si128();
    // 2^12 loop iters might overflow 16-bit counter, so
    // cap it at 2^11 iters per chunk
    const int maxiters = 2048;
    while (N > maxiters) {
        popcount_sse2_16_helper_2x1(bufR0+base,bufR1+base,bufB+base,maxiters,count32_R0,count32_R1);
        base += maxiters*4;
        N -= maxiters;
    }
    if (N > 0) popcount_sse2_16_helper_2x1(bufR0+base,bufR1+base,bufB+base,N,count32_R0,count32_R1);


    // Partial horizontal add-reduction
    __m128i partial = _mm_add_epi32(_mm_unpackhi_epi32(count32_R0,count32_R1),
                                    _mm_unpacklo_epi32(count32_R0,count32_R1));
    _mm_store_si128((__m128i*)tmp,partial);
    counts[0] = tmp[0] + tmp[2];
    counts[1] = tmp[1] + tmp[3];

    return;
}

inline int popcount_1fp(unsigned* buf,const int nwords) {
    assert(nwords % 16 == 0);
    return popcount_sse2_16_1(buf,nwords);
}
inline int popcount_2fp(unsigned* bufA,unsigned* bufB,const int nwords) {
    assert(nwords % 16 == 0);
    return popcount_sse2_16_2(bufA,bufB,nwords);
}
inline void popcount_2fp_2( unsigned* bufR0, unsigned* bufR1, unsigned* bufB, int nwords,int* counts) {
    assert(nwords % 8 == 0);
    popcount_sse2_16_2x1(bufR0,bufR1,bufB,nwords,counts);
    return;
}
