/**************************************************************************

 Written by Imran S. Haque (ihaque@cs.stanford.edu)

 Copyright (c) 2011 Stanford University.
 All rights reserved.

 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.
     * Neither the name of Stanford University nor the names of its contributors
       may be used to endorse or promote products derived from this software without
       specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

**************************************************************************/

#ifdef _DEBUG
#include <assert.h>
#else
#define assert(x)
#endif

#include <emmintrin.h>
#include <stdint.h>

static inline __m128i  _mm_swizzle_epi32_zwxy(__m128i reg) {  return _mm_shuffle_epi32(reg,0x4E); }
inline __m128i _mm_srli1_epi8(__m128i v) {
    // Avg returns (a_i + b_i + 1) >> 1; so make b_i = -1 to compute (a >> 1)
    const __m128i neg1 = _mm_set1_epi8(0xFF);
    return _mm_avg_epu8(v,neg1);
}
inline __m128i _mm_srli2_epi8(__m128i v) {
    // Faster than mask-and-shift approach used in srli4
    return _mm_srli1_epi8(_mm_srli1_epi8(v));
}
inline __m128i _mm_srli4_epi8(__m128i v) {
    // Using ANDNOT lets us share the magic number with what's used for the popcount itself
    const __m128i mask = _mm_set1_epi32(0x0F0F0F0F);
    return _mm_srli_epi16(_mm_andnot_si128(mask,v),4);
}
inline __m128i popcount_sse2_8_helper_1(unsigned* buf, int N) {
    __m128i* vbuf = (__m128i*)buf;
    __m128i total,count0,count1,count2,count3;
    int i;
    __m128i B0 = _mm_set1_epi32(0x55555555);
    __m128i B1 = _mm_set1_epi32(0x33333333);
    __m128i B2 = _mm_set1_epi32(0x0F0F0F0F);
    total = _mm_setzero_si128();
    for (i = 0; i < N; i+=4) {
        __m128i v0 = _mm_load_si128(vbuf+i+0);
        __m128i v1 = _mm_load_si128(vbuf+i+1);
        __m128i v2 = _mm_load_si128(vbuf+i+2);
        __m128i v3 = _mm_load_si128(vbuf+i+3);
        
        count0 = _mm_sub_epi8(v0, _mm_and_si128(B0, _mm_srli1_epi8(v0)));
        count1 = _mm_sub_epi8(v1, _mm_and_si128(B0, _mm_srli1_epi8(v1)));
        count2 = _mm_sub_epi8(v2, _mm_and_si128(B0, _mm_srli1_epi8(v2)));
        count3 = _mm_sub_epi8(v3, _mm_and_si128(B0, _mm_srli1_epi8(v3)));

        count0 = _mm_add_epi8(_mm_and_si128(B1,count0),
                               _mm_and_si128(B1,_mm_srli2_epi8(count0)));
        count1 = _mm_add_epi8(_mm_and_si128(B1,count1),
                               _mm_and_si128(B1,_mm_srli2_epi8(count1)));
        count2 = _mm_add_epi8(_mm_and_si128(B1,count2),
                               _mm_and_si128(B1,_mm_srli2_epi8(count2)));
        count3 = _mm_add_epi8(_mm_and_si128(B1,count3),
                               _mm_and_si128(B1,_mm_srli2_epi8(count3)));

        count0 = _mm_and_si128(B2,_mm_add_epi8(count0,
                                               _mm_srli4_epi8(count0)));
        count1 = _mm_and_si128(B2,_mm_add_epi8(count1,
                                               _mm_srli4_epi8(count1)));
        count2 = _mm_and_si128(B2,_mm_add_epi8(count2,
                                               _mm_srli4_epi8(count2)));
        count3 = _mm_and_si128(B2,_mm_add_epi8(count3,
                                               _mm_srli4_epi8(count3)));

        total = _mm_add_epi8(total,_mm_add_epi8(_mm_add_epi8(count0,count1),
                                                _mm_add_epi8(count2,count3)));

    }

    // Reduce 16*8b->{-,-,-,16b,-,-,-,16b}
    const __m128i ZERO = _mm_setzero_si128();
    return _mm_sad_epu8(total,ZERO);
}
inline __m128i popcount_sse2_8_helper_2(unsigned* bufA, unsigned* bufB, int N) {
    __m128i* vbufA = (__m128i*)bufA;
    __m128i* vbufB = (__m128i*)bufB;
    __m128i total,count0,count1,count2,count3;
    int i;
    __m128i B0 = _mm_set1_epi32(0x55555555);
    __m128i B1 = _mm_set1_epi32(0x33333333);
    __m128i B2 = _mm_set1_epi32(0x0F0F0F0F);
    total = _mm_setzero_si128();
    for (i = 0; i < N; i+=4) {
        __m128i v0 = _mm_and_si128(_mm_load_si128(vbufA+i+0),_mm_load_si128(vbufB+i+0));
        __m128i v1 = _mm_and_si128(_mm_load_si128(vbufA+i+1),_mm_load_si128(vbufB+i+1));
        __m128i v2 = _mm_and_si128(_mm_load_si128(vbufA+i+2),_mm_load_si128(vbufB+i+2));
        __m128i v3 = _mm_and_si128(_mm_load_si128(vbufA+i+3),_mm_load_si128(vbufB+i+3));
        
        count0 = _mm_sub_epi8(v0, _mm_and_si128(B0, _mm_srli1_epi8(v0)));
        count1 = _mm_sub_epi8(v1, _mm_and_si128(B0, _mm_srli1_epi8(v1)));
        count2 = _mm_sub_epi8(v2, _mm_and_si128(B0, _mm_srli1_epi8(v2)));
        count3 = _mm_sub_epi8(v3, _mm_and_si128(B0, _mm_srli1_epi8(v3)));

        count0 = _mm_add_epi8(_mm_and_si128(B1,count0),
                               _mm_and_si128(B1,_mm_srli2_epi8(count0)));
        count1 = _mm_add_epi8(_mm_and_si128(B1,count1),
                               _mm_and_si128(B1,_mm_srli2_epi8(count1)));
        count2 = _mm_add_epi8(_mm_and_si128(B1,count2),
                               _mm_and_si128(B1,_mm_srli2_epi8(count2)));
        count3 = _mm_add_epi8(_mm_and_si128(B1,count3),
                               _mm_and_si128(B1,_mm_srli2_epi8(count3)));

        count0 = _mm_and_si128(B2,_mm_add_epi8(count0,
                                               _mm_srli4_epi8(count0)));
        count1 = _mm_and_si128(B2,_mm_add_epi8(count1,
                                               _mm_srli4_epi8(count1)));
        count2 = _mm_and_si128(B2,_mm_add_epi8(count2,
                                               _mm_srli4_epi8(count2)));
        count3 = _mm_and_si128(B2,_mm_add_epi8(count3,
                                               _mm_srli4_epi8(count3)));

        total = _mm_add_epi8(total,_mm_add_epi8(_mm_add_epi8(count0,count1),
                                                _mm_add_epi8(count2,count3)));

    }

    // Reduce 16*8b->{-,-,-,16b,-,-,-,16b}
    const __m128i ZERO = _mm_setzero_si128();
    return _mm_sad_epu8(total,ZERO);
}
inline void popcount_sse2_8_helper_2x1(const unsigned* bufR0, const unsigned* bufR1, const unsigned* bufB, const int N,__m128i& countR0,__m128i& countR1) {
    __m128i* vbufR0 = (__m128i*)bufR0;
    __m128i* vbufR1 = (__m128i*)bufR1;
    __m128i* vbufB = (__m128i*)bufB;
    __m128i totalR0,totalR1,countR0_0,countR0_1,countR1_0,countR1_1;
    const __m128i M0 = _mm_set1_epi32(0x55555555);
    const __m128i M1 = _mm_set1_epi32(0x33333333);
    const __m128i M2 = _mm_set1_epi32(0x0F0F0F0F);
    __m128i vB0,vB1;
    totalR0 = _mm_setzero_si128();
    totalR1 = _mm_setzero_si128();
    for (int i = 0; i < N; i+=2) {
            vB0 = _mm_load_si128(vbufB+i);
            vB1 = _mm_load_si128(vbufB+i+1);
            __m128i vR0_0 = _mm_and_si128(_mm_load_si128(vbufR0+i+0),vB0);
            __m128i vR0_1 = _mm_and_si128(_mm_load_si128(vbufR0+i+1),vB1);
            __m128i vR1_0 = _mm_and_si128(_mm_load_si128(vbufR1+i+0),vB0);
            __m128i vR1_1 = _mm_and_si128(_mm_load_si128(vbufR1+i+1),vB1);
            
            countR0_0 = _mm_sub_epi8(vR0_0, _mm_and_si128(M0, _mm_srli1_epi8(vR0_0)));
            countR0_1 = _mm_sub_epi8(vR0_1, _mm_and_si128(M0, _mm_srli1_epi8(vR0_1)));
            countR1_0 = _mm_sub_epi8(vR1_0, _mm_and_si128(M0, _mm_srli1_epi8(vR1_0)));
            countR1_1 = _mm_sub_epi8(vR1_1, _mm_and_si128(M0, _mm_srli1_epi8(vR1_1)));

            countR0_0 = _mm_add_epi8(_mm_and_si128(M1,countR0_0),
                                      _mm_and_si128(M1,_mm_srli2_epi8(countR0_0)));
            countR0_1 = _mm_add_epi8(_mm_and_si128(M1,countR0_1),
                                      _mm_and_si128(M1,_mm_srli2_epi8(countR0_1)));
            countR1_0 = _mm_add_epi8(_mm_and_si128(M1,countR1_0),
                                      _mm_and_si128(M1,_mm_srli2_epi8(countR1_0)));
            countR1_1 = _mm_add_epi8(_mm_and_si128(M1,countR1_1),
                                      _mm_and_si128(M1,_mm_srli2_epi8(countR1_1)));

            countR0_0 = _mm_and_si128(M2,_mm_add_epi8(countR0_0,
                                                       _mm_srli4_epi8(countR0_0)));
            countR0_1 = _mm_and_si128(M2,_mm_add_epi8(countR0_1,
                                                       _mm_srli4_epi8(countR0_1)));
            countR1_0 = _mm_and_si128(M2,_mm_add_epi8(countR1_0,
                                                       _mm_srli4_epi8(countR1_0)));
            countR1_1 = _mm_and_si128(M2,_mm_add_epi8(countR1_1,
                                                       _mm_srli4_epi8(countR1_1)));

            totalR0 = _mm_add_epi8(totalR0,_mm_add_epi8(countR0_0,countR0_1));
            totalR1 = _mm_add_epi8(totalR1,_mm_add_epi8(countR1_0,countR1_1));
    }
    // Reduce 16*8b->{-,-,-,16b,-,-,-,16b}
    const __m128i ZERO = _mm_setzero_si128();
    countR0 = _mm_add_epi32(countR0,_mm_sad_epu8(totalR0,ZERO));
    countR1 = _mm_add_epi32(countR1,_mm_sad_epu8(totalR1,ZERO));
    return;
}
inline int popcount_sse2_8_1(unsigned* buf,int n) {
    int N = n/4;
    __m128i count32 = _mm_setzero_si128();
    // 2^5 loop iters might overflow 8-bit counter, so
    // cap it at 2^4 iters per chunk
    const int inner_maxits = 16;
    while (N > inner_maxits) {
        count32 = _mm_add_epi32(count32,popcount_sse2_8_helper_1(buf,inner_maxits));
        buf += inner_maxits*4;
        N -= inner_maxits;
    }
    if (N > 0) count32 = _mm_add_epi32(count32,popcount_sse2_8_helper_1(buf,N));

    // Layout coming from PSADBW accumulation is 2*{0,32}: 0 S1 0 S0
    int count;
    _mm_store_ss((float*)&count,(__m128)(_mm_add_epi32(count32,_mm_shuffle_epi32(count32,_MM_SHUFFLE(2,2,2,2)))));
    return count;

}
inline int popcount_sse2_8_2(unsigned* bufA,unsigned* bufB,int n) {
    int N = n/4;
    __m128i count32 = _mm_setzero_si128();
    // 2^5 loop iters might overflow 8-bit counter, so
    // cap it at 2^4 iters per chunk
    const int inner_maxits = 16;
    while (N > inner_maxits) {
        count32 = _mm_add_epi32(count32,popcount_sse2_8_helper_2(bufA,bufB,inner_maxits));
        bufA += inner_maxits*4;
        bufB += inner_maxits*4;
        N -= inner_maxits;
    }
    if (N > 0) count32 = _mm_add_epi32(count32,popcount_sse2_8_helper_2(bufA,bufB,N));

    // Layout coming from PSADBW accumulation is 2*{0,32}: 0 S1 0 S0
    int count;
    _mm_store_ss((float*)&count,(__m128)(_mm_add_epi32(count32,_mm_shuffle_epi32(count32,_MM_SHUFFLE(2,2,2,2)))));
    return count;

}
inline void popcount_sse2_8_2x1(unsigned* bufR0,unsigned* bufR1,unsigned* bufB,int n,int* counts) {
    int N = n/4;
    __m128i count32R0 = _mm_setzero_si128();
    __m128i count32R1 = _mm_setzero_si128();
    // 2^5 loop iters might overflow 8-bit counter, so
    // cap it at 2^4 iters per chunk
    const int inner_maxits = 16;
    while (N > inner_maxits) {
        popcount_sse2_8_helper_2x1(bufR0,bufR1,bufB,inner_maxits,count32R0,count32R1);
        bufR0 += inner_maxits*4;
        bufR1 += inner_maxits*4;
        bufB += inner_maxits*4;
        N -= inner_maxits;
    }
    if (N > 0) popcount_sse2_8_helper_2x1(bufR0,bufR1,bufB,inner_maxits,count32R0,count32R1);

    // Layout coming from PSADBW accumulation is 2*{0,32}: 0 S1 0 S0
    // countR0: 0 S01 0 S00 (hi --> lo)
    // countR1: 0 S11 0 S10
    __m128i t1 = _mm_slli_si128(count32R1,4); // S11 0 S10 0
    __m128i t2 = _mm_or_si128(count32R0,t1); // S11 S01 S10 S00
    __m128i t3 = _mm_swizzle_epi32_zwxy(t2); // S10 S00 S11 S01
    __m128i final = _mm_add_epi32(t2,t3); // S1 S0 S1 S0
    _mm_storel_epi64((__m128i*)counts,final);
    return;
}

inline int popcount_1fp(unsigned* buf,const int nwords) {
    assert(nwords % 16 == 0);
    return popcount_sse2_8_1(buf,nwords);
}
inline int popcount_2fp(unsigned* bufA,unsigned* bufB,const int nwords) {
    assert(nwords % 16 == 0);
    return popcount_sse2_8_2(bufA,bufB,nwords);
}
inline void popcount_2fp_2( unsigned* bufR0, unsigned* bufR1, unsigned* bufB, int nwords,int* counts) {
    assert(nwords % 8 == 0);
    popcount_sse2_8_2x1(bufR0,bufR1,bufB,nwords,counts);
    return;
}
