/**************************************************************************

 Written by Imran S. Haque (ihaque@cs.stanford.edu)

 Copyright (c) 2011 Stanford University.
 All rights reserved.

 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.
     * Neither the name of Stanford University nor the names of its contributors
       may be used to endorse or promote products derived from this software without
       specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

**************************************************************************/

#include <nmmintrin.h>
#include <stdint.h>
#ifdef _DEBUG
#include <assert.h>
#else
#define assert(x) 
#endif

inline int popcount_1fp(unsigned* buf,const int nwords) {
    int count = 0;
    assert(nwords % 4 == 0);
    int nquads = nwords/2;
    uint64_t* buf64 = (uint64_t*)buf;
    for (int i = 0; i < nquads; i += 2) {
        count += _mm_popcnt_u64(buf64[i]) + _mm_popcnt_u64(buf64[i+1]);
    }
    return count;
}
inline int popcount_2fp(unsigned* bufA,unsigned* bufB,const int nwords) {
    int count = 0;
    assert(nwords % 8 == 0);
    int nquads = nwords/2;
    uint64_t* a64 = (uint64_t*)bufA;
    uint64_t* b64 = (uint64_t*)bufB;
    for (int i = 0; i < nquads; i += 4) {
        count +=  _mm_popcnt_u64(a64[i]&b64[i])     + _mm_popcnt_u64(a64[i+1]&b64[i+1])
                + _mm_popcnt_u64(a64[i+2]&b64[i+2]) + _mm_popcnt_u64(a64[i+3]&b64[i+3]);
    }
    return count;
}
inline void popcount_2fp_2( unsigned* bufR0, unsigned* bufR1, unsigned* bufB, int nwords,int* counts) {
    assert(nwords % 8 == 0);
    counts[0] = counts[1] = 0;
    int nquads = nwords/2;
    uint64_t* r0 = (uint64_t*)bufR0;
    uint64_t* r1 = (uint64_t*)bufR1;
    uint64_t* b64 = (uint64_t*)bufB;
    uint64_t b0,b1;
    for (int i = 0; i < nquads; i += 2) {
        b0 = b64[i]; b1 = b64[i+1];
        counts[0] +=  _mm_popcnt_u64(r0[i  ]&b0) + _mm_popcnt_u64(r0[i+1]&b1);
        counts[1] +=  _mm_popcnt_u64(r1[i]  &b0) + _mm_popcnt_u64(r1[i+1]&b1);
    }
    return;
}
