/**************************************************************************

 Written by Imran S. Haque (ihaque@cs.stanford.edu)

 Copyright (c) 2011 Stanford University.
 All rights reserved.

 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.
     * Neither the name of Stanford University nor the names of its contributors
       may be used to endorse or promote products derived from this software without
       specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

**************************************************************************/

#include <iostream>
#include <fstream>
#include <string>
#include <bitset>
#include <sstream>
#include <vector>
#include <cstdlib>

#include <string.h>
#include <math.h>

#include <ctime>
#include <sys/time.h>
#include <omp.h>
static inline double getTimeMilliseconds(void) {
    struct timeval tv;
    gettimeofday(&tv,NULL);
    return tv.tv_sec*1000.0 + tv.tv_usec/1000.0;
}


#include "nbrCounts.h"

using namespace std;

int main(int argc, const char **argv) {

    if (argc != 5) {
        cerr << "usage: " << (string)argv[0] << " database.fps vendor.fps cutoff outfile" << endl;
        exit(0);
    }

    float cutoff = atof(argv[3]);
    
    int numgpus;
    cudaGetDeviceCount(&numgpus);
    #pragma omp parallel
    {
        // Schedule threads onto GPUs round-robin
        cudaSetDevice(omp_get_thread_num()%numgpus);
    }

    NbrCounts bmk(argv[1],argv[2],cutoff,argv[4]);
    //bmk.run_cuda(); //return the numToFind id_minSims vector
    //bmk.run_cpu_sse4_omp_T(); //return the numToFind id_minSims vector
    // run the GPU version
    double start = getTimeMilliseconds();
    double gpuEnd = getTimeMilliseconds();
    // run the CPU version
    //bmk.run_cpu(); //return the numToFind id_minSims vector
    double cpuEnd = getTimeMilliseconds();
    cout << "CPU    runtime: "<< (cpuEnd-gpuEnd) <<"ms"<<endl;
    
    //bmk.run_cpu_sse4(); //return the numToFind id_minSims vector
    double sseEnd = getTimeMilliseconds();
    cout << "SSE    runtime: "<< (sseEnd-cpuEnd) <<"ms"<<endl;

    //bmk.run_cpu_sse4_omp(); //return the numToFind id_minSims vector
    double ompEnd = getTimeMilliseconds();
    cout << "OMP    runtime: "<< (ompEnd-sseEnd) <<"ms"<<endl;

    bmk.run_hybrid(); //return the numToFind id_minSims vector
    double hybridEnd = getTimeMilliseconds();
    cout << "HYB    runtime: "<< (hybridEnd-ompEnd) <<"ms"<<endl;

    //bmk.run_cpu_sse4_omp_T(); //return the numToFind id_minSims vector
    double ompTEnd = getTimeMilliseconds();
    cout << "OMPT   runtime: "<< (ompTEnd-hybridEnd) <<"ms"<<endl;

    bmk.run_cuda(); //return the numToFind id_minSims vector
    double cudaEnd = getTimeMilliseconds();
    cout << "GPU    runtime: "<< (cudaEnd-ompTEnd) <<"ms"<<endl;

    bmk.run_cuda_unroll4(); //return the numToFind id_minSims vector
    double cuda4End = getTimeMilliseconds();
    cout << "GPU4   runtime: "<< (cuda4End-cudaEnd) <<"ms"<<endl;

    bmk.run_cpu_sse4_omp_mag_block();
    double nvcbEnd = getTimeMilliseconds();
    cout << "NVCB   runtime: "<< (nvcbEnd-cuda4End) <<"ms"<<endl;

    bmk.read_fps();
    bmk.read_fps();
    bmk.read_fps();
    bmk.read_fps();
    bmk.read_fps();
    double fpEnd = getTimeMilliseconds();
    cout << "READFP runtime: "<< (fpEnd-nvcbEnd)/5 <<"ms"<<endl;

    return 0;
}

