/**************************************************************************

 Written by Imran S. Haque (ihaque@cs.stanford.edu) and Trevor Kramer

 Copyright (c) 2011 Stanford University.
 All rights reserved.

 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.
     * Neither the name of Stanford University nor the names of its contributors
       may be used to endorse or promote products derived from this software without
       specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

**************************************************************************/

#ifndef __BENCHMARK__
#define __BENCHMARK__

#include <iostream>
#include <fstream>
#include <string>
#include <bitset>
#include <sstream>
#include <vector>

#include <string.h>
#include <math.h>

using namespace std;

int bitcount_32bit(unsigned int n);
string int32_toString(unsigned int i);
void printBitMatrix_padded(unsigned int* bitMatrix, int numFpsPadded);
unsigned int char32BitString_to_uint32(const char *char32);
void compute_magnitudes(const unsigned *qfps,unsigned* magQs,const int numqfps,const int fpT_stride_ints);
class NbrCounts {

public:
    int numFpsTrue;
    int numFpsTrue_vendor;
    int numFpsPadded;
    int fpBitPadLen;
    int numThreads;
    int fpSizeInts;
    string databasefile;
    string vendorfile;
    string resultsfile;
    vector<string> ids;
    vector<string> ids_vendor;
    unsigned int* fingerprints;
    unsigned int* fingerprintsT_alloc;
    unsigned int* fingerprintsT;
    unsigned int* magD;
    double * thresh_partD;
    unsigned int fpT_stride_ints;
    unsigned int* fingerprints_vendor; //NOT edge on
    float cutoff;
    bool verbose_gpu_matrix;
    bool verbose_gpu_looping;

    //NbrCounts(string databasefile, string vendorfile, float cutoff, string resultsfile);
    NbrCounts(string databasefile, string vendorfile, float cutoff, string resultsfile);
    unsigned int myScreen(unsigned int* fp);
    unsigned int myScreen_sse4(unsigned int* fp);
    unsigned int myScreen_sse4_omp(unsigned int* fp);
    unsigned int myScreen_sse4_omp_T(unsigned int* fp);
    unsigned int myScreen_sse4_omp_vecT(unsigned int* fp);
    unsigned int myScreen_sse4_omp_vecT_mag(unsigned int* fp);
    unsigned int myScreen_sse4_omp_mag(unsigned int* fp);
    void myScreen_sse4_omp_vecT_mag_block(unsigned int* qfps,const int numqfps,unsigned int* counts);
    void myScreen_sse4_omp_mag_block(unsigned int* qfps,const int numqfps,unsigned int* counts,unsigned int* magQ=NULL);
    void myScreen_cuda(unsigned int* qfps,const int numqfps,unsigned int* counts);
    void myScreen_cuda_unroll4(unsigned int* qfps,const int numqfps,unsigned int* counts);
    void myScreen_hybrid(unsigned int* qfps,const int numqfps,unsigned int* counts);
    void run_cpu();
    void read_fps();
    void run_cpu_sse4();
    void run_cpu_sse4_omp();
    void run_cpu_sse4_omp_T();
    void run_cuda();
    void run_hybrid();
    void run_cuda_unroll4();
    void run_cpu_sse4_omp_vecT();
    void run_cpu_sse4_omp_vecT_mag();
    void run_cpu_sse4_omp_mag();
    void run_cpu_sse4_omp_vecT_mag_block();
    void run_cpu_sse4_omp_mag_block();
    ~NbrCounts();
};

struct gpuData {
    unsigned int *d_dfps, *d_qfps, *d_qmags, *d_dmags, *d_counts;
    unsigned int *h_dfps, *h_qfps, *h_qmags, *h_dmags, *h_counts;
    int numdbfps, numqfps;
    int fpT_stride_ints;
};
void hybrid_helper_start(struct gpuData& data,double cutoff);
void hybrid_helper_finish(struct gpuData& data);
#endif
